#pragma once
#include "IR_config.h"
#include "RingBuffer.h"

// #define IRDEBUG

#ifdef IRDEBUG
#define wrHigh PA1  // Запись HIGH инициирована                     // green
#define wrLow PA0   // Запись LOW инициирована                      // blue
#define writeOp PA5 // Операция записи, 1 пульс для 0 и 2 для 1     // orange
// Исправленные ошибки                                              // purle
// 1 пульс: fix
#define errOut PA4
#define up PA3
#define down PA2
#endif

/////////////////////////////////////////////////////////////////////////////////////////////////

#define riseTime riseSyncTime              //* bitTime */  893U        // TODO: Должно высчитываться медианой
#define riseTolerance tolerance /* 250U */ // погрешность
#define riseTimeMax (riseTime + riseTolerance)
#define riseTimeMin (riseTime - riseTolerance)
#define aroundRise(t) (riseTimeMin < t && t < riseTimeMax)
#define IR_timeout (riseTimeMax * (8 + syncBits + 1)) // us // таймаут в 8 data + 3 sync + 1

class IR_Encoder;
class IR_DecoderRaw : virtual public IR_FOX
{
    friend IR_Encoder;

protected:
    PackInfo packInfo;
    IR_Encoder *encoder; // Указатель на парный передатчик
    bool availableRaw();

public:
    //////////////////////////////////////////////////////////////////////////
    /// @brief Конструктор
    /// @param pin Номер вывода прерывания/данных от приёмника (2 или 3 для atmega 328p)
    /// @param addr Адрес приёмника
    /// @param encPair Указатель на передатчик, работающий в паре
    IR_DecoderRaw(const uint8_t pin, uint16_t addr, IR_Encoder *encPair = nullptr);

    void isr();  // Функция прерывания
    void tick(); // Обработка приёмника, необходима для работы

    inline bool isOverflow() { return isBufferOverflow; }; // Буффер переполнился
    bool isSubOverflow();
    inline bool isReciving() { return isBufferOverflow; }; // Возвращает true, если происходит приём пакета

    //////////////////////////////////////////////////////////////////////////
private:
    ErrorsStruct errors;
    bool isAvailable = false;
    uint16_t packSize;
    uint16_t crcValue;
    volatile uint16_t isPairSending = 0; // Флаг передачи парного передатчика
    volatile bool isRecive = false;      // Флаг приёма
    volatile bool isPreamb = false;      // флаг начальной последовости
    volatile bool isSubBufferOverflow = false;
    bool isBufferOverflow = false; // Флаг переполнения буффера данных
    bool isWrongPack = false;      // Флаг битого пакета

    uint16_t riseSyncTime = bitTime; // Подстраиваемое время бита в мкс

    ////////////////////////////////////////////////////////////////////////
    volatile uint32_t currentSubBufferIndex; // Счетчик текущей позиции во вспомогательном буфере фронтов/спадов

    struct FrontStorage
    {                               // Структура для хранения времени и направления фронта/спада
        volatile uint32_t time = 0; // Время
        volatile bool dir = false;  // Направление (true = ↑; false = ↓)
        // volatile FrontStorage *next = nullptr; // Указатель на следующий связанный фронт/спад, или nullptr если конец
    };
    volatile FrontStorage *lastFront = nullptr;           // Указатель последнего фронта/спада
    volatile FrontStorage *firstUnHandledFront = nullptr; // Указатель первого необработанного фронта/спада
    // volatile FrontStorage subBuffer[subBufferSize];       // вспомогательный буфер для хранения необработанных фронтов/спадов

    RingBuffer<FrontStorage, subBufferSize> subBuffer;

    ////////////////////////////////////////////////////////////////////////
    uint8_t dataBuffer[dataByteSizeMax]{0};                           // Буффер данных
    volatile uint32_t prevRise, prevPrevRise, prevFall, prevPrevFall; // Время предыдущих фронтов/спадов

    volatile uint32_t risePeriod;
    volatile uint32_t highTime;
    volatile uint32_t lowTime;

    uint32_t oldTime;
    uint16_t wrongCounter;

    int8_t highCount;
    int8_t lowCount;
    int8_t allCount;

    uint16_t errorCounter = 0;     // Счётчик ошибок
    int8_t preambFrontCounter = 0; // Счётчик __/``` ↑ преамбулы
    int16_t bufBitPos = 0;         // Позиция для записи бита в буффер

private:
    void listenStart(); // @brief Слушатель для работы isReciving()

    /// @brief Проверка CRC. Проверяет len байт со значением crc, пришедшим в пакете
    /// @param len Длина в байтах проверяемых данных
    /// @param crc Результат рассчёта crc (Выходной параметр)
    /// @return true если crc верно
    bool crcCheck(uint8_t len, uint16_t &crc);

    ////////////////////////////////////////////////////////////////////////
    bool isData = true;                  // Флаг относится ли бит к данным, или битам синхронизации
    uint16_t i_dataBuffer;               // Счётчик буфера данных
    uint8_t nextControlBit = bitPerByte; // Метка для смены флага isData
    uint8_t i_syncBit;                   // Счётчик битов синхронизации
    uint8_t err_syncBit;                 // Счётчик ошибок синхронизации

    /// @brief Запиь бита в буффер, а так же проверка битов синхранизации и их фильтрация
    /// @param Бит данных
    void writeToBuffer(bool);
    ////////////////////////////////////////////////////////////////////////

    void firstRX(); /// @brief Установка и сброс начальных значений и флагов в готовность к приёму данных

    /// @brief Целочисленное деление с округлением вверх
    /// @param val Значение
    /// @param divider Делитель
    /// @return Результат
    uint16_t ceil_div(uint16_t val, uint16_t divider);

#ifdef IRDEBUG
    uint32_t wrCounter;
    inline void errPulse(uint8_t pin, uint8_t count);
    inline void infoPulse(uint8_t pin, uint8_t count);
#endif
};
